# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_DDPM_v2.ipynb.

# %% auto 0
__all__ = ['MixedPrecision', 'AccelerateCB']

# %% ../nbs/10_DDPM_v2.ipynb 3
import pickle,gzip,math,os,time,shutil,torch,random,logging
import fastcore.all as fc,matplotlib as mpl,numpy as np,matplotlib.pyplot as plt
from collections.abc import Mapping
from pathlib import Path
from functools import partial

from fastcore.foundation import L
import torchvision.transforms.functional as TF,torch.nn.functional as F
from torch import tensor,nn,optim
from torch.utils.data import DataLoader,default_collate
from torch.nn import init
from torch.optim import lr_scheduler

from .datasets import *
from .conv import *
from .learner import *
from .activations import *
from .init import *
from .sgd import *
from .resnet import *
from .augment import *

# %% ../nbs/10_DDPM_v2.ipynb 41
class MixedPrecision(TrainCB):
    order = DeviceCB.order+10
    
    def before_fit(self, learn): self.scaler = torch.cuda.amp.GradScaler()

    def before_batch(self, learn):
        self.autocast = torch.autocast("cuda", dtype=torch.float16)
        self.autocast.__enter__()

    def after_loss(self, learn): self.autocast.__exit__(None, None, None)
        
    def backward(self, learn): self.scaler.scale(learn.loss).backward()

    def step(self, learn):
        self.scaler.step(learn.opt)
        self.scaler.update()

# %% ../nbs/10_DDPM_v2.ipynb 49
from accelerate import Accelerator

# %% ../nbs/10_DDPM_v2.ipynb 50
class AccelerateCB(TrainCB):
    order = DeviceCB.order+10
    def __init__(self, n_inp=1, mixed_precision="fp16"):
        super().__init__(n_inp=n_inp)
        self.acc = Accelerator(mixed_precision=mixed_precision)
        
    def before_fit(self, learn):
        learn.model,learn.opt,learn.dls.train,learn.dls.valid = self.acc.prepare(
            learn.model, learn.opt, learn.dls.train, learn.dls.valid)

    def backward(self, learn): self.acc.backward(learn.loss)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_learner.ipynb.

# %% auto 0
__all__ = ['CancelFitException', 'CancelBatchException', 'CancelEpochException', 'Callback', 'run_cbs', 'SingleBatchCB', 'to_cpu',
           'MetricsCB', 'DeviceCB', 'TrainCB', 'ProgressCB', 'with_cbs', 'Learner', 'TrainLearner', 'MomentumLearner',
           'LRFinderCB', 'lr_find', 'WandbCallback', 'MLflowCallback']

# %% ../nbs/04_learner.ipynb 1
import math,torch,matplotlib.pyplot as plt
import fastcore.all as fc
from collections.abc import Mapping
from operator import attrgetter
from functools import partial
from copy import copy
import wandb
import mlflow
import mlflow.pytorch
from pathlib import Path
from torch import optim
import torch.nn.functional as F
from .conv import *
from fastprogress import progress_bar,master_bar

# %% ../nbs/04_learner.ipynb 14
class CancelFitException(Exception): pass
class CancelBatchException(Exception): pass
class CancelEpochException(Exception): pass

# %% ../nbs/04_learner.ipynb 15
class Callback(): order = 0

# %% ../nbs/04_learner.ipynb 16
def run_cbs(cbs, method_nm, learn=None):
    for cb in sorted(cbs, key=attrgetter('order')):
        method = getattr(cb, method_nm, None)
        if method is not None: method(learn)

# %% ../nbs/04_learner.ipynb 22
class SingleBatchCB(Callback):
    order = 1
    def after_batch(self, learn): raise CancelFitException()

# %% ../nbs/04_learner.ipynb 31
from torcheval.metrics import MulticlassAccuracy,Mean

# %% ../nbs/04_learner.ipynb 34
def to_cpu(x):
    if isinstance(x, Mapping): return {k:to_cpu(v) for k,v in x.items()}
    if isinstance(x, list): return [to_cpu(o) for o in x]
    if isinstance(x, tuple): return tuple(to_cpu(list(x)))
    res = x.detach().cpu()
    return res.float() if res.dtype==torch.float16 else res

# %% ../nbs/04_learner.ipynb 35
class MetricsCB(Callback):
    def __init__(self, *ms, **metrics):
        for o in ms: metrics[type(o).__name__] = o
        self.metrics = metrics
        self.all_metrics = copy(metrics)
        self.all_metrics['loss'] = self.loss = Mean()

    def _log(self, d): print(d)
    def before_fit(self, learn): learn.metrics = self
    def before_epoch(self, learn): [o.reset() for o in self.all_metrics.values()]

    def after_epoch(self, learn):
        log = {k:f'{v.compute():.3f}' for k,v in self.all_metrics.items()}
        log['epoch'] = learn.epoch
        log['train'] = 'train' if learn.model.training else 'eval'
        self._log(log)

    def after_batch(self, learn):
        x,y,*_ = to_cpu(learn.batch)
        for m in self.metrics.values(): m.update(to_cpu(learn.preds), y)
        self.loss.update(to_cpu(learn.loss), weight=len(x))

# %% ../nbs/04_learner.ipynb 36
class DeviceCB(Callback):
    def __init__(self, device=def_device): fc.store_attr()
    def before_fit(self, learn):
        if hasattr(learn.model, 'to'): learn.model.to(self.device)
    def before_batch(self, learn): learn.batch = to_device(learn.batch, device=self.device)

# %% ../nbs/04_learner.ipynb 40
class TrainCB(Callback):
    def __init__(self, n_inp=1): self.n_inp = n_inp
    def predict(self, learn): learn.preds = learn.model(*learn.batch[:self.n_inp])
    def get_loss(self, learn): learn.loss = learn.loss_func(learn.preds, *learn.batch[self.n_inp:])
    def backward(self, learn): learn.loss.backward()
    def step(self, learn): learn.opt.step()
    def zero_grad(self, learn): learn.opt.zero_grad()

# %% ../nbs/04_learner.ipynb 42
class ProgressCB(Callback):
    order = MetricsCB.order+1
    def __init__(self, plot=False): self.plot = plot
    def before_fit(self, learn):
        learn.epochs = self.mbar = master_bar(learn.epochs)
        self.first = True
        if hasattr(learn, 'metrics'): learn.metrics._log = self._log
        self.losses = []
        self.val_losses = []

    def _log(self, d):
        if self.first:
            self.mbar.write(list(d), table=True)
            self.first = False
        self.mbar.write(list(d.values()), table=True)

    def before_epoch(self, learn): learn.dl = progress_bar(learn.dl, leave=False, parent=self.mbar)
    def after_batch(self, learn):
        learn.dl.comment = f'{learn.loss:.3f}'
        if self.plot and hasattr(learn, 'metrics') and learn.training:
            self.losses.append(learn.loss.item())
            if self.val_losses: self.mbar.update_graph([[fc.L.range(self.losses), self.losses],[fc.L.range(learn.epoch).map(lambda x: (x+1)*len(learn.dls.train)), self.val_losses]])
    
    def after_epoch(self, learn): 
        if not learn.training:
            if self.plot and hasattr(learn, 'metrics'): 
                self.val_losses.append(learn.metrics.all_metrics['loss'].compute())
                self.mbar.update_graph([[fc.L.range(self.losses), self.losses],[fc.L.range(learn.epoch+1).map(lambda x: (x+1)*len(learn.dls.train)), self.val_losses]])

# %% ../nbs/04_learner.ipynb 48
class with_cbs:
    def __init__(self, nm): self.nm = nm
    def __call__(self, f):
        def _f(o, *args, **kwargs):
            try:
                o.callback(f'before_{self.nm}')
                f(o, *args, **kwargs)
                o.callback(f'after_{self.nm}')
            except globals()[f'Cancel{self.nm.title()}Exception']: pass
            finally: o.callback(f'cleanup_{self.nm}')
        return _f

# %% ../nbs/04_learner.ipynb 49
class Learner():
    def __init__(self, model, dls=(0,), loss_func=F.mse_loss, lr=0.1, cbs=None, opt_func=optim.Adam):
        cbs = fc.L(cbs)
        fc.store_attr()

    @with_cbs('batch')
    def _one_batch(self):
        self.predict()
        self.callback('after_predict')
        self.get_loss()
        self.callback('after_loss')
        if self.training:
            self.backward()
            self.callback('after_backward')
            self.step()
            self.callback('after_step')
            self.zero_grad()

    @with_cbs('epoch')
    def _one_epoch(self):
        for self.iter,self.batch in enumerate(self.dl): self._one_batch()

    def one_epoch(self, training):
        self.model.train(training)
        self.dl = self.dls.train if training else self.dls.valid
        self._one_epoch()

    @with_cbs('fit')
    def _fit(self, train, valid):
        for self.epoch in self.epochs:
            if train: self.one_epoch(True)
            if valid: torch.no_grad()(self.one_epoch)(False)

    def fit(self, n_epochs=1, train=True, valid=True, cbs=None, lr=None):
        cbs = fc.L(cbs)
        # `add_cb` and `rm_cb` were added in lesson 18
        for cb in cbs: self.cbs.append(cb)
        try:
            self.n_epochs = n_epochs
            self.epochs = range(n_epochs)
            if lr is None: lr = self.lr
            if self.opt_func: self.opt = self.opt_func(self.model.parameters(), lr)
            self._fit(train, valid)
        finally:
            for cb in cbs: self.cbs.remove(cb)

    def __getattr__(self, name):
        if name in ('predict','get_loss','backward','step','zero_grad'): return partial(self.callback, name)
        raise AttributeError(name)

    def callback(self, method_nm): run_cbs(self.cbs, method_nm, self)
    
    @property
    def training(self): return self.model.training

# %% ../nbs/04_learner.ipynb 52
class TrainLearner(Learner):
    def predict(self): self.preds = self.model(self.batch[0])
    def get_loss(self): self.loss = self.loss_func(self.preds, self.batch[1])
    def backward(self): self.loss.backward()
    def step(self): self.opt.step()
    def zero_grad(self): self.opt.zero_grad()

# %% ../nbs/04_learner.ipynb 53
class MomentumLearner(TrainLearner):
    def __init__(self, model, dls, loss_func, lr=None, cbs=None, opt_func=optim.SGD, mom=0.85):
        self.mom = mom
        super().__init__(model, dls, loss_func, lr, cbs, opt_func)

    def zero_grad(self):
        with torch.no_grad():
            for p in self.model.parameters(): p.grad *= self.mom

# %% ../nbs/04_learner.ipynb 58
from torch.optim.lr_scheduler import ExponentialLR

# %% ../nbs/04_learner.ipynb 60
class LRFinderCB(Callback):
    def __init__(self, gamma=1.3, max_mult=3): fc.store_attr()
    
    def before_fit(self, learn):
        self.sched = ExponentialLR(learn.opt, self.gamma)
        self.lrs,self.losses = [],[]
        self.min = math.inf

    def after_batch(self, learn):
        if not learn.training: raise CancelEpochException()
        self.lrs.append(learn.opt.param_groups[0]['lr'])
        loss = to_cpu(learn.loss)
        self.losses.append(loss)
        if loss < self.min: self.min = loss
        if math.isnan(loss) or (loss > self.min*self.max_mult):
            raise CancelFitException()
        self.sched.step()

    def cleanup_fit(self, learn):
        plt.plot(self.lrs, self.losses)
        plt.xscale('log')

# %% ../nbs/04_learner.ipynb 62
@fc.patch
def lr_find(self:Learner, gamma=1.3, max_mult=3, start_lr=1e-5, max_epochs=10):
    self.fit(max_epochs, lr=start_lr, cbs=LRFinderCB(gamma=gamma, max_mult=max_mult))

# %% ../nbs/04_learner.ipynb 65
class WandbCallback(Callback):
    def __init__(
        self,
        project_name: str,
        run_name: str = None,
        config: dict = None,
        log_model: bool = True,
        log_frequency: int = 100,
        save_best_model: bool = True,
        monitor: str = 'val_loss',
        mode: str = 'min',
        save_model_checkpoint: bool = True,
        checkpoint_dir: str = './models',
        log_gradients: bool = False,
        log_preds: bool = False,
        preds_frequency: int = 500,
    ):
        """
        Initializes the WandbCallback.

        Args:
            project_name (str): Name of the W&B project.
            run_name (str, optional): Name of the W&B run. Defaults to None.
            config (dict, optional): Hyperparameters and configurations. Defaults to None.
            log_model (bool, optional): Log model architecture to W&B. Defaults to True.
            log_frequency (int, optional): Frequency (in steps) to log training metrics. Defaults to 100.
            save_best_model (bool, optional): Save the best model during training. Defaults to True.
            monitor (str, optional): Metric to monitor for best model saving. Defaults to 'val_loss'.
            mode (str, optional): 'min' or 'max' to minimize or maximize the monitored metric. Defaults to 'min'.
            save_model_checkpoint (bool, optional): Save model checkpoint after each epoch. Defaults to True.
            checkpoint_dir (str, optional): Directory to save model checkpoints. Defaults to './models'.
            log_gradients (bool, optional): Log gradients histograms. Defaults to False.
            log_preds (bool, optional): Log model predictions. Defaults to False.
            preds_frequency (int, optional): Frequency (in steps) to log predictions. Defaults to 500.
        """
        fc.store_attr()

    def before_fit(self, learn):
        # Initialize W&B run
        self.run = wandb.init(project=self.project_name, name=self.run_name, config=self.config)
        self.best_metric = None
        self.operator = torch.lt if self.mode == 'min' else torch.gt
        self.checkpoint_path = Path(self.checkpoint_dir)
        self.checkpoint_path.mkdir(parents=True, exist_ok=True)
        
        if self.log_model:
            # Log model architecture
            wandb.watch(learn.model, log='all' if self.log_gradients else 'parameters')

    def after_batch(self, learn):
        if learn.training and (learn.iter % self.log_frequency == 0):
            metrics = {
                'train/loss': learn.loss.item(),
                'train/epoch': learn.epoch + (learn.iter / len(learn.dl))
            }
            # Log metrics to W&B
            wandb.log(metrics, step=learn.iter_total)
            
            if self.log_gradients:
                # Log gradients
                for name, param in learn.model.named_parameters():
                    if param.grad is not None:
                        wandb.log({f'gradients/{name}': wandb.Histogram(param.grad.cpu().numpy())}, step=learn.iter_total)
            
            if self.log_preds and (learn.iter % self.preds_frequency == 0):
                # Log predictions (assuming classification task)
                inputs, targets = learn.batch[:2]
                preds = torch.argmax(learn.preds, dim=1)
                table = wandb.Table(columns=["input", "prediction", "target"])
                for inp, pred, target in zip(inputs.cpu(), preds.cpu(), targets.cpu()):
                    table.add_data(wandb.Image(inp), pred.item(), target.item())
                wandb.log({"predictions": table}, step=learn.iter_total)

    def after_epoch(self, learn):
        # Compute validation metrics
        val_metrics = {f'val/{k}': v.compute().item() for k, v in learn.metrics.metrics.items()}
        val_metrics['val/loss'] = learn.metrics.loss.compute().item()
        val_metrics['epoch'] = learn.epoch
        # Log validation metrics
        wandb.log(val_metrics, step=learn.iter_total)
        
        # Save model checkpoint
        if self.save_model_checkpoint:
            epoch_checkpoint_path = self.checkpoint_path / f'model_epoch_{learn.epoch}.pth'
            torch.save(learn.model.state_dict(), epoch_checkpoint_path)
            wandb.save(str(epoch_checkpoint_path))
        
        # Save best model
        current_metric = val_metrics.get(f'val/{self.monitor}', val_metrics.get('val/loss'))
        if self.save_best_model and current_metric is not None:
            if self.best_metric is None or self.operator(current_metric, self.best_metric):
                self.best_metric = current_metric
                best_checkpoint_path = self.checkpoint_path / 'best_model.pth'
                torch.save(learn.model.state_dict(), best_checkpoint_path)
                wandb.save(str(best_checkpoint_path))
                wandb.run.summary[f'best_{self.monitor}'] = self.best_metric

    def after_fit(self, learn):
        # Finish W&B run
        wandb.finish()




class MLflowCallback(Callback):
    def __init__(
        self,
        experiment_name: str,
        run_name: str = None,
        tracking_uri: str = None,
        config: dict = None,
        log_model: bool = True,
        log_frequency: int = 100,
        save_best_model: bool = True,
        monitor: str = 'val_loss',
        mode: str = 'min',
        save_model_checkpoint: bool = True,
        checkpoint_dir: str = './models',
    ):
        """
        Initializes the MLflowCallback.

        Args:
            experiment_name (str): Name of the MLflow experiment.
            run_name (str, optional): Name of the MLflow run. Defaults to None.
            tracking_uri (str, optional): URI of the tracking server. Defaults to None (local server).
            config (dict, optional): Hyperparameters and configurations. Defaults to None.
            log_model (bool, optional): Log model architecture to MLflow. Defaults to True.
            log_frequency (int, optional): Frequency (in steps) to log training metrics. Defaults to 100.
            save_best_model (bool, optional): Save the best model during training. Defaults to True.
            monitor (str, optional): Metric to monitor for best model saving. Defaults to 'val_loss'.
            mode (str, optional): 'min' or 'max' to minimize or maximize the monitored metric. Defaults to 'min'.
            save_model_checkpoint (bool, optional): Save model checkpoint after each epoch. Defaults to True.
            checkpoint_dir (str, optional): Directory to save model checkpoints. Defaults to './models'.
        """
        fc.store_attr()

    def before_fit(self, learn):
        # Set MLflow tracking URI if provided
        if self.tracking_uri:
            mlflow.set_tracking_uri(self.tracking_uri)
        
        # Set experiment and run
        mlflow.set_experiment(self.experiment_name)
        self.run = mlflow.start_run(run_name=self.run_name)
        
        # Log configuration parameters
        if self.config:
            mlflow.log_params(self.config)
        
        # Prepare for saving checkpoints
        self.best_metric = None
        self.operator = torch.lt if self.mode == 'min' else torch.gt
        self.checkpoint_path = Path(self.checkpoint_dir)
        self.checkpoint_path.mkdir(parents=True, exist_ok=True)
        
        # Log model architecture if needed
        if self.log_model:
            mlflow.pytorch.log_model(learn.model, "model_architecture")

    def after_batch(self, learn):
        if learn.training and (learn.iter % self.log_frequency == 0):
            metrics = {
                'train/loss': learn.loss.item(),
                'train/epoch': learn.epoch + (learn.iter / len(learn.dl))
            }
            mlflow.log_metrics(metrics, step=learn.iter_total)

    def after_epoch(self, learn):
        # Compute validation metrics
        val_metrics = {f'val/{k}': v.compute().item() for k, v in learn.metrics.metrics.items()}
        val_metrics['val/loss'] = learn.metrics.loss.compute().item()
        val_metrics['epoch'] = learn.epoch
        
        # Log validation metrics
        mlflow.log_metrics(val_metrics, step=learn.iter_total)
        
        # Save model checkpoint
        if self.save_model_checkpoint:
            epoch_checkpoint_path = self.checkpoint_path / f'model_epoch_{learn.epoch}.pth'
            torch.save(learn.model.state_dict(), epoch_checkpoint_path)
            mlflow.log_artifact(str(epoch_checkpoint_path))
        
        # Save best model
        current_metric = val_metrics.get(f'val/{self.monitor}', val_metrics.get('val/loss'))
        if self.save_best_model and current_metric is not None:
            if self.best_metric is None or self.operator(current_metric, self.best_metric):
                self.best_metric = current_metric
                best_checkpoint_path = self.checkpoint_path / 'best_model.pth'
                torch.save(learn.model.state_dict(), best_checkpoint_path)
                mlflow.log_artifact(str(best_checkpoint_path))
                mlflow.log_metric(f'best_{self.monitor}', self.best_metric, step=learn.iter_total)

    def after_fit(self, learn):
        # Finish MLflow run
        mlflow.end_run()



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/rapidai',
                'doc_host': 'https://m-gopichand.github.io',
                'git_url': 'https://github.com/m-gopichand/rapidai',
                'lib_path': 'rapidai'},
  'syms': { 'rapidai.accel': { 'rapidai.accel.AccelerateCB': ('ddpm_v2.html#acceleratecb', 'rapidai/accel.py'),
                               'rapidai.accel.AccelerateCB.__init__': ('ddpm_v2.html#acceleratecb.__init__', 'rapidai/accel.py'),
                               'rapidai.accel.AccelerateCB.backward': ('ddpm_v2.html#acceleratecb.backward', 'rapidai/accel.py'),
                               'rapidai.accel.AccelerateCB.before_fit': ('ddpm_v2.html#acceleratecb.before_fit', 'rapidai/accel.py'),
                               'rapidai.accel.MixedPrecision': ('ddpm_v2.html#mixedprecision', 'rapidai/accel.py'),
                               'rapidai.accel.MixedPrecision.after_loss': ('ddpm_v2.html#mixedprecision.after_loss', 'rapidai/accel.py'),
                               'rapidai.accel.MixedPrecision.backward': ('ddpm_v2.html#mixedprecision.backward', 'rapidai/accel.py'),
                               'rapidai.accel.MixedPrecision.before_batch': ( 'ddpm_v2.html#mixedprecision.before_batch',
                                                                              'rapidai/accel.py'),
                               'rapidai.accel.MixedPrecision.before_fit': ('ddpm_v2.html#mixedprecision.before_fit', 'rapidai/accel.py'),
                               'rapidai.accel.MixedPrecision.step': ('ddpm_v2.html#mixedprecision.step', 'rapidai/accel.py')},
            'rapidai.activations': { 'rapidai.activations.ActivationStats': ('activations.html#activationstats', 'rapidai/activations.py'),
                                     'rapidai.activations.ActivationStats.__init__': ( 'activations.html#activationstats.__init__',
                                                                                       'rapidai/activations.py'),
                                     'rapidai.activations.ActivationStats.color_dim': ( 'activations.html#activationstats.color_dim',
                                                                                        'rapidai/activations.py'),
                                     'rapidai.activations.ActivationStats.dead_chart': ( 'activations.html#activationstats.dead_chart',
                                                                                         'rapidai/activations.py'),
                                     'rapidai.activations.ActivationStats.plot_stats': ( 'activations.html#activationstats.plot_stats',
                                                                                         'rapidai/activations.py'),
                                     'rapidai.activations.Hook': ('activations.html#hook', 'rapidai/activations.py'),
                                     'rapidai.activations.Hook.__del__': ('activations.html#hook.__del__', 'rapidai/activations.py'),
                                     'rapidai.activations.Hook.__init__': ('activations.html#hook.__init__', 'rapidai/activations.py'),
                                     'rapidai.activations.Hook.remove': ('activations.html#hook.remove', 'rapidai/activations.py'),
                                     'rapidai.activations.Hooks': ('activations.html#hooks', 'rapidai/activations.py'),
                                     'rapidai.activations.Hooks.__del__': ('activations.html#hooks.__del__', 'rapidai/activations.py'),
                                     'rapidai.activations.Hooks.__delitem__': ( 'activations.html#hooks.__delitem__',
                                                                                'rapidai/activations.py'),
                                     'rapidai.activations.Hooks.__enter__': ('activations.html#hooks.__enter__', 'rapidai/activations.py'),
                                     'rapidai.activations.Hooks.__exit__': ('activations.html#hooks.__exit__', 'rapidai/activations.py'),
                                     'rapidai.activations.Hooks.__init__': ('activations.html#hooks.__init__', 'rapidai/activations.py'),
                                     'rapidai.activations.Hooks.remove': ('activations.html#hooks.remove', 'rapidai/activations.py'),
                                     'rapidai.activations.HooksCallback': ('activations.html#hookscallback', 'rapidai/activations.py'),
                                     'rapidai.activations.HooksCallback.__init__': ( 'activations.html#hookscallback.__init__',
                                                                                     'rapidai/activations.py'),
                                     'rapidai.activations.HooksCallback.__iter__': ( 'activations.html#hookscallback.__iter__',
                                                                                     'rapidai/activations.py'),
                                     'rapidai.activations.HooksCallback.__len__': ( 'activations.html#hookscallback.__len__',
                                                                                    'rapidai/activations.py'),
                                     'rapidai.activations.HooksCallback._hookfunc': ( 'activations.html#hookscallback._hookfunc',
                                                                                      'rapidai/activations.py'),
                                     'rapidai.activations.HooksCallback.after_fit': ( 'activations.html#hookscallback.after_fit',
                                                                                      'rapidai/activations.py'),
                                     'rapidai.activations.HooksCallback.before_fit': ( 'activations.html#hookscallback.before_fit',
                                                                                       'rapidai/activations.py'),
                                     'rapidai.activations.append_stats': ('activations.html#append_stats', 'rapidai/activations.py'),
                                     'rapidai.activations.get_hist': ('activations.html#get_hist', 'rapidai/activations.py'),
                                     'rapidai.activations.get_min': ('activations.html#get_min', 'rapidai/activations.py'),
                                     'rapidai.activations.set_seed': ('activations.html#set_seed', 'rapidai/activations.py')},
            'rapidai.augment': { 'rapidai.augment.CapturePreds': ('augment.html#capturepreds', 'rapidai/augment.py'),
                                 'rapidai.augment.CapturePreds.after_batch': ( 'augment.html#capturepreds.after_batch',
                                                                               'rapidai/augment.py'),
                                 'rapidai.augment.CapturePreds.after_fit': ('augment.html#capturepreds.after_fit', 'rapidai/augment.py'),
                                 'rapidai.augment.CapturePreds.before_fit': ('augment.html#capturepreds.before_fit', 'rapidai/augment.py'),
                                 'rapidai.augment.RandCopy': ('augment.html#randcopy', 'rapidai/augment.py'),
                                 'rapidai.augment.RandCopy.__init__': ('augment.html#randcopy.__init__', 'rapidai/augment.py'),
                                 'rapidai.augment.RandCopy.forward': ('augment.html#randcopy.forward', 'rapidai/augment.py'),
                                 'rapidai.augment.RandErase': ('augment.html#randerase', 'rapidai/augment.py'),
                                 'rapidai.augment.RandErase.__init__': ('augment.html#randerase.__init__', 'rapidai/augment.py'),
                                 'rapidai.augment.RandErase.forward': ('augment.html#randerase.forward', 'rapidai/augment.py'),
                                 'rapidai.augment._flops': ('augment.html#_flops', 'rapidai/augment.py'),
                                 'rapidai.augment._rand_copy1': ('augment.html#_rand_copy1', 'rapidai/augment.py'),
                                 'rapidai.augment._rand_erase1': ('augment.html#_rand_erase1', 'rapidai/augment.py'),
                                 'rapidai.augment.capture_preds': ('augment.html#capture_preds', 'rapidai/augment.py'),
                                 'rapidai.augment.rand_copy': ('augment.html#rand_copy', 'rapidai/augment.py'),
                                 'rapidai.augment.rand_erase': ('augment.html#rand_erase', 'rapidai/augment.py'),
                                 'rapidai.augment.show_image_batch': ('augment.html#show_image_batch', 'rapidai/augment.py'),
                                 'rapidai.augment.summary': ('augment.html#summary', 'rapidai/augment.py')},
            'rapidai.conv': { 'rapidai.conv.collate_device': ('convolutions.html#collate_device', 'rapidai/conv.py'),
                              'rapidai.conv.conv': ('convolutions.html#conv', 'rapidai/conv.py'),
                              'rapidai.conv.device': ('convolutions.html#device', 'rapidai/conv.py'),
                              'rapidai.conv.to_device': ('convolutions.html#to_device', 'rapidai/conv.py')},
            'rapidai.core': {'rapidai.core.foo': ('core.html#foo', 'rapidai/core.py')},
            'rapidai.datasets': { 'rapidai.datasets.DataLoaders': ('datasets.html#dataloaders', 'rapidai/datasets.py'),
                                  'rapidai.datasets.DataLoaders.__init__': ('datasets.html#dataloaders.__init__', 'rapidai/datasets.py'),
                                  'rapidai.datasets.DataLoaders.from_dd': ('datasets.html#dataloaders.from_dd', 'rapidai/datasets.py'),
                                  'rapidai.datasets.collate_dict': ('datasets.html#collate_dict', 'rapidai/datasets.py'),
                                  'rapidai.datasets.get_grid': ('datasets.html#get_grid', 'rapidai/datasets.py'),
                                  'rapidai.datasets.inplace': ('datasets.html#inplace', 'rapidai/datasets.py'),
                                  'rapidai.datasets.show_image': ('datasets.html#show_image', 'rapidai/datasets.py'),
                                  'rapidai.datasets.show_images': ('datasets.html#show_images', 'rapidai/datasets.py'),
                                  'rapidai.datasets.subplots': ('datasets.html#subplots', 'rapidai/datasets.py')},
            'rapidai.diffusion': { 'rapidai.diffusion.DownBlock': ('diffusion-attn-cond.html#downblock', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.DownBlock.__init__': ( 'diffusion-attn-cond.html#downblock.__init__',
                                                                             'rapidai/diffusion.py'),
                                   'rapidai.diffusion.DownBlock.forward': ( 'diffusion-attn-cond.html#downblock.forward',
                                                                            'rapidai/diffusion.py'),
                                   'rapidai.diffusion.EmbResBlock': ('diffusion-attn-cond.html#embresblock', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.EmbResBlock.__init__': ( 'diffusion-attn-cond.html#embresblock.__init__',
                                                                               'rapidai/diffusion.py'),
                                   'rapidai.diffusion.EmbResBlock.forward': ( 'diffusion-attn-cond.html#embresblock.forward',
                                                                              'rapidai/diffusion.py'),
                                   'rapidai.diffusion.EmbUNetModel': ('diffusion-attn-cond.html#embunetmodel', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.EmbUNetModel.__init__': ( 'diffusion-attn-cond.html#embunetmodel.__init__',
                                                                                'rapidai/diffusion.py'),
                                   'rapidai.diffusion.EmbUNetModel.forward': ( 'diffusion-attn-cond.html#embunetmodel.forward',
                                                                               'rapidai/diffusion.py'),
                                   'rapidai.diffusion.SelfAttention': ('diffusion-attn-cond.html#selfattention', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.SelfAttention.__init__': ( 'diffusion-attn-cond.html#selfattention.__init__',
                                                                                 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.SelfAttention.forward': ( 'diffusion-attn-cond.html#selfattention.forward',
                                                                                'rapidai/diffusion.py'),
                                   'rapidai.diffusion.SelfAttention2D': ( 'diffusion-attn-cond.html#selfattention2d',
                                                                          'rapidai/diffusion.py'),
                                   'rapidai.diffusion.SelfAttention2D.forward': ( 'diffusion-attn-cond.html#selfattention2d.forward',
                                                                                  'rapidai/diffusion.py'),
                                   'rapidai.diffusion.UpBlock': ('diffusion-attn-cond.html#upblock', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.UpBlock.__init__': ( 'diffusion-attn-cond.html#upblock.__init__',
                                                                           'rapidai/diffusion.py'),
                                   'rapidai.diffusion.UpBlock.forward': ( 'diffusion-attn-cond.html#upblock.forward',
                                                                          'rapidai/diffusion.py'),
                                   'rapidai.diffusion.abar': ('diffusion-attn-cond.html#abar', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.collate_ddpm': ('diffusion-attn-cond.html#collate_ddpm', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.cond_sample': ('diffusion-attn-cond.html#cond_sample', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.ddim_step': ('diffusion-attn-cond.html#ddim_step', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.dl_ddpm': ('diffusion-attn-cond.html#dl_ddpm', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.inv_abar': ('diffusion-attn-cond.html#inv_abar', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.lin': ('diffusion-attn-cond.html#lin', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.noisify': ('diffusion-attn-cond.html#noisify', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.pre_conv': ('diffusion-attn-cond.html#pre_conv', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.sample': ('diffusion-attn-cond.html#sample', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.saved': ('diffusion-attn-cond.html#saved', 'rapidai/diffusion.py'),
                                   'rapidai.diffusion.timestep_embedding': ( 'diffusion-attn-cond.html#timestep_embedding',
                                                                             'rapidai/diffusion.py'),
                                   'rapidai.diffusion.upsample': ('diffusion-attn-cond.html#upsample', 'rapidai/diffusion.py')},
            'rapidai.fid': { 'rapidai.fid.ImageEval': ('fid.html#imageeval', 'rapidai/fid.py'),
                             'rapidai.fid.ImageEval.__init__': ('fid.html#imageeval.__init__', 'rapidai/fid.py'),
                             'rapidai.fid.ImageEval.fid': ('fid.html#imageeval.fid', 'rapidai/fid.py'),
                             'rapidai.fid.ImageEval.get_feats': ('fid.html#imageeval.get_feats', 'rapidai/fid.py'),
                             'rapidai.fid.ImageEval.kid': ('fid.html#imageeval.kid', 'rapidai/fid.py'),
                             'rapidai.fid._calc_fid': ('fid.html#_calc_fid', 'rapidai/fid.py'),
                             'rapidai.fid._calc_kid': ('fid.html#_calc_kid', 'rapidai/fid.py'),
                             'rapidai.fid._calc_stats': ('fid.html#_calc_stats', 'rapidai/fid.py'),
                             'rapidai.fid._sqrtm_newton_schulz': ('fid.html#_sqrtm_newton_schulz', 'rapidai/fid.py'),
                             'rapidai.fid._squared_mmd': ('fid.html#_squared_mmd', 'rapidai/fid.py')},
            'rapidai.imports': {},
            'rapidai.init': { 'rapidai.init.BatchTransformCB': ('initializing.html#batchtransformcb', 'rapidai/init.py'),
                              'rapidai.init.BatchTransformCB.__init__': ('initializing.html#batchtransformcb.__init__', 'rapidai/init.py'),
                              'rapidai.init.BatchTransformCB.before_batch': ( 'initializing.html#batchtransformcb.before_batch',
                                                                              'rapidai/init.py'),
                              'rapidai.init.GeneralRelu': ('initializing.html#generalrelu', 'rapidai/init.py'),
                              'rapidai.init.GeneralRelu.__init__': ('initializing.html#generalrelu.__init__', 'rapidai/init.py'),
                              'rapidai.init.GeneralRelu.forward': ('initializing.html#generalrelu.forward', 'rapidai/init.py'),
                              'rapidai.init._lsuv_stats': ('initializing.html#_lsuv_stats', 'rapidai/init.py'),
                              'rapidai.init.clean_ipython_hist': ('initializing.html#clean_ipython_hist', 'rapidai/init.py'),
                              'rapidai.init.clean_mem': ('initializing.html#clean_mem', 'rapidai/init.py'),
                              'rapidai.init.clean_tb': ('initializing.html#clean_tb', 'rapidai/init.py'),
                              'rapidai.init.conv': ('initializing.html#conv', 'rapidai/init.py'),
                              'rapidai.init.get_model': ('initializing.html#get_model', 'rapidai/init.py'),
                              'rapidai.init.init_weights': ('initializing.html#init_weights', 'rapidai/init.py'),
                              'rapidai.init.lsuv_init': ('initializing.html#lsuv_init', 'rapidai/init.py'),
                              'rapidai.init.plot_func': ('initializing.html#plot_func', 'rapidai/init.py')},
            'rapidai.learner': { 'rapidai.learner.Callback': ('learner.html#callback', 'rapidai/learner.py'),
                                 'rapidai.learner.CancelBatchException': ('learner.html#cancelbatchexception', 'rapidai/learner.py'),
                                 'rapidai.learner.CancelEpochException': ('learner.html#cancelepochexception', 'rapidai/learner.py'),
                                 'rapidai.learner.CancelFitException': ('learner.html#cancelfitexception', 'rapidai/learner.py'),
                                 'rapidai.learner.DeviceCB': ('learner.html#devicecb', 'rapidai/learner.py'),
                                 'rapidai.learner.DeviceCB.__init__': ('learner.html#devicecb.__init__', 'rapidai/learner.py'),
                                 'rapidai.learner.DeviceCB.before_batch': ('learner.html#devicecb.before_batch', 'rapidai/learner.py'),
                                 'rapidai.learner.DeviceCB.before_fit': ('learner.html#devicecb.before_fit', 'rapidai/learner.py'),
                                 'rapidai.learner.LRFinderCB': ('learner.html#lrfindercb', 'rapidai/learner.py'),
                                 'rapidai.learner.LRFinderCB.__init__': ('learner.html#lrfindercb.__init__', 'rapidai/learner.py'),
                                 'rapidai.learner.LRFinderCB.after_batch': ('learner.html#lrfindercb.after_batch', 'rapidai/learner.py'),
                                 'rapidai.learner.LRFinderCB.before_fit': ('learner.html#lrfindercb.before_fit', 'rapidai/learner.py'),
                                 'rapidai.learner.LRFinderCB.cleanup_fit': ('learner.html#lrfindercb.cleanup_fit', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner': ('learner.html#learner', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner.__getattr__': ('learner.html#learner.__getattr__', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner.__init__': ('learner.html#learner.__init__', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner._fit': ('learner.html#learner._fit', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner._one_batch': ('learner.html#learner._one_batch', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner._one_epoch': ('learner.html#learner._one_epoch', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner.callback': ('learner.html#learner.callback', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner.fit': ('learner.html#learner.fit', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner.one_epoch': ('learner.html#learner.one_epoch', 'rapidai/learner.py'),
                                 'rapidai.learner.Learner.training': ('learner.html#learner.training', 'rapidai/learner.py'),
                                 'rapidai.learner.MLflowCallback': ('learner.html#mlflowcallback', 'rapidai/learner.py'),
                                 'rapidai.learner.MLflowCallback.__init__': ('learner.html#mlflowcallback.__init__', 'rapidai/learner.py'),
                                 'rapidai.learner.MLflowCallback.after_batch': ( 'learner.html#mlflowcallback.after_batch',
                                                                                 'rapidai/learner.py'),
                                 'rapidai.learner.MLflowCallback.after_epoch': ( 'learner.html#mlflowcallback.after_epoch',
                                                                                 'rapidai/learner.py'),
                                 'rapidai.learner.MLflowCallback.after_fit': ( 'learner.html#mlflowcallback.after_fit',
                                                                               'rapidai/learner.py'),
                                 'rapidai.learner.MLflowCallback.before_fit': ( 'learner.html#mlflowcallback.before_fit',
                                                                                'rapidai/learner.py'),
                                 'rapidai.learner.MetricsCB': ('learner.html#metricscb', 'rapidai/learner.py'),
                                 'rapidai.learner.MetricsCB.__init__': ('learner.html#metricscb.__init__', 'rapidai/learner.py'),
                                 'rapidai.learner.MetricsCB._log': ('learner.html#metricscb._log', 'rapidai/learner.py'),
                                 'rapidai.learner.MetricsCB.after_batch': ('learner.html#metricscb.after_batch', 'rapidai/learner.py'),
                                 'rapidai.learner.MetricsCB.after_epoch': ('learner.html#metricscb.after_epoch', 'rapidai/learner.py'),
                                 'rapidai.learner.MetricsCB.before_epoch': ('learner.html#metricscb.before_epoch', 'rapidai/learner.py'),
                                 'rapidai.learner.MetricsCB.before_fit': ('learner.html#metricscb.before_fit', 'rapidai/learner.py'),
                                 'rapidai.learner.MomentumLearner': ('learner.html#momentumlearner', 'rapidai/learner.py'),
                                 'rapidai.learner.MomentumLearner.__init__': ( 'learner.html#momentumlearner.__init__',
                                                                               'rapidai/learner.py'),
                                 'rapidai.learner.MomentumLearner.zero_grad': ( 'learner.html#momentumlearner.zero_grad',
                                                                                'rapidai/learner.py'),
                                 'rapidai.learner.ProgressCB': ('learner.html#progresscb', 'rapidai/learner.py'),
                                 'rapidai.learner.ProgressCB.__init__': ('learner.html#progresscb.__init__', 'rapidai/learner.py'),
                                 'rapidai.learner.ProgressCB._log': ('learner.html#progresscb._log', 'rapidai/learner.py'),
                                 'rapidai.learner.ProgressCB.after_batch': ('learner.html#progresscb.after_batch', 'rapidai/learner.py'),
                                 'rapidai.learner.ProgressCB.after_epoch': ('learner.html#progresscb.after_epoch', 'rapidai/learner.py'),
                                 'rapidai.learner.ProgressCB.before_epoch': ('learner.html#progresscb.before_epoch', 'rapidai/learner.py'),
                                 'rapidai.learner.ProgressCB.before_fit': ('learner.html#progresscb.before_fit', 'rapidai/learner.py'),
                                 'rapidai.learner.SingleBatchCB': ('learner.html#singlebatchcb', 'rapidai/learner.py'),
                                 'rapidai.learner.SingleBatchCB.after_batch': ( 'learner.html#singlebatchcb.after_batch',
                                                                                'rapidai/learner.py'),
                                 'rapidai.learner.TrainCB': ('learner.html#traincb', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainCB.__init__': ('learner.html#traincb.__init__', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainCB.backward': ('learner.html#traincb.backward', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainCB.get_loss': ('learner.html#traincb.get_loss', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainCB.predict': ('learner.html#traincb.predict', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainCB.step': ('learner.html#traincb.step', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainCB.zero_grad': ('learner.html#traincb.zero_grad', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainLearner': ('learner.html#trainlearner', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainLearner.backward': ('learner.html#trainlearner.backward', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainLearner.get_loss': ('learner.html#trainlearner.get_loss', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainLearner.predict': ('learner.html#trainlearner.predict', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainLearner.step': ('learner.html#trainlearner.step', 'rapidai/learner.py'),
                                 'rapidai.learner.TrainLearner.zero_grad': ('learner.html#trainlearner.zero_grad', 'rapidai/learner.py'),
                                 'rapidai.learner.WandbCallback': ('learner.html#wandbcallback', 'rapidai/learner.py'),
                                 'rapidai.learner.WandbCallback.__init__': ('learner.html#wandbcallback.__init__', 'rapidai/learner.py'),
                                 'rapidai.learner.WandbCallback.after_batch': ( 'learner.html#wandbcallback.after_batch',
                                                                                'rapidai/learner.py'),
                                 'rapidai.learner.WandbCallback.after_epoch': ( 'learner.html#wandbcallback.after_epoch',
                                                                                'rapidai/learner.py'),
                                 'rapidai.learner.WandbCallback.after_fit': ('learner.html#wandbcallback.after_fit', 'rapidai/learner.py'),
                                 'rapidai.learner.WandbCallback.before_fit': ( 'learner.html#wandbcallback.before_fit',
                                                                               'rapidai/learner.py'),
                                 'rapidai.learner.lr_find': ('learner.html#lr_find', 'rapidai/learner.py'),
                                 'rapidai.learner.run_cbs': ('learner.html#run_cbs', 'rapidai/learner.py'),
                                 'rapidai.learner.to_cpu': ('learner.html#to_cpu', 'rapidai/learner.py'),
                                 'rapidai.learner.with_cbs': ('learner.html#with_cbs', 'rapidai/learner.py'),
                                 'rapidai.learner.with_cbs.__call__': ('learner.html#with_cbs.__call__', 'rapidai/learner.py'),
                                 'rapidai.learner.with_cbs.__init__': ('learner.html#with_cbs.__init__', 'rapidai/learner.py')},
            'rapidai.resnet': { 'rapidai.resnet.ResBlock': ('resnet.html#resblock', 'rapidai/resnet.py'),
                                'rapidai.resnet.ResBlock.__init__': ('resnet.html#resblock.__init__', 'rapidai/resnet.py'),
                                'rapidai.resnet.ResBlock.forward': ('resnet.html#resblock.forward', 'rapidai/resnet.py'),
                                'rapidai.resnet._conv_block': ('resnet.html#_conv_block', 'rapidai/resnet.py')},
            'rapidai.sgd': { 'rapidai.sgd.BaseSchedCB': ('accel_sgd.html#baseschedcb', 'rapidai/sgd.py'),
                             'rapidai.sgd.BaseSchedCB.__init__': ('accel_sgd.html#baseschedcb.__init__', 'rapidai/sgd.py'),
                             'rapidai.sgd.BaseSchedCB._step': ('accel_sgd.html#baseschedcb._step', 'rapidai/sgd.py'),
                             'rapidai.sgd.BaseSchedCB.before_fit': ('accel_sgd.html#baseschedcb.before_fit', 'rapidai/sgd.py'),
                             'rapidai.sgd.BatchSchedCB': ('accel_sgd.html#batchschedcb', 'rapidai/sgd.py'),
                             'rapidai.sgd.BatchSchedCB.after_batch': ('accel_sgd.html#batchschedcb.after_batch', 'rapidai/sgd.py'),
                             'rapidai.sgd.EpochSchedCB': ('accel_sgd.html#epochschedcb', 'rapidai/sgd.py'),
                             'rapidai.sgd.EpochSchedCB.after_epoch': ('accel_sgd.html#epochschedcb.after_epoch', 'rapidai/sgd.py'),
                             'rapidai.sgd.HasLearnCB': ('accel_sgd.html#haslearncb', 'rapidai/sgd.py'),
                             'rapidai.sgd.HasLearnCB.after_fit': ('accel_sgd.html#haslearncb.after_fit', 'rapidai/sgd.py'),
                             'rapidai.sgd.HasLearnCB.before_fit': ('accel_sgd.html#haslearncb.before_fit', 'rapidai/sgd.py'),
                             'rapidai.sgd.RecorderCB': ('accel_sgd.html#recordercb', 'rapidai/sgd.py'),
                             'rapidai.sgd.RecorderCB.__init__': ('accel_sgd.html#recordercb.__init__', 'rapidai/sgd.py'),
                             'rapidai.sgd.RecorderCB.after_batch': ('accel_sgd.html#recordercb.after_batch', 'rapidai/sgd.py'),
                             'rapidai.sgd.RecorderCB.before_fit': ('accel_sgd.html#recordercb.before_fit', 'rapidai/sgd.py'),
                             'rapidai.sgd.RecorderCB.plot': ('accel_sgd.html#recordercb.plot', 'rapidai/sgd.py')},
            'rapidai.training': { 'rapidai.training.Dataset': ('minibatch_training.html#dataset', 'rapidai/training.py'),
                                  'rapidai.training.Dataset.__getitem__': ( 'minibatch_training.html#dataset.__getitem__',
                                                                            'rapidai/training.py'),
                                  'rapidai.training.Dataset.__init__': ('minibatch_training.html#dataset.__init__', 'rapidai/training.py'),
                                  'rapidai.training.Dataset.__len__': ('minibatch_training.html#dataset.__len__', 'rapidai/training.py'),
                                  'rapidai.training.accuracy': ('minibatch_training.html#accuracy', 'rapidai/training.py'),
                                  'rapidai.training.fit': ('minibatch_training.html#fit', 'rapidai/training.py'),
                                  'rapidai.training.get_dls': ('minibatch_training.html#get_dls', 'rapidai/training.py'),
                                  'rapidai.training.report': ('minibatch_training.html#report', 'rapidai/training.py')}}}



import torch, random, datasets, math, fastcore.all as fc, numpy as np, matplotlib as mpl, matplotlib.pyplot as plt
import torchvision.transforms as T
import torchvision.transforms.functional as TF,torch.nn.functional as F

from torch.utils.data import DataLoader,default_collate
from pathlib import Path
from torch import nn,tensor
from torch.nn import init
from fastcore.foundation import L
from datasets import load_dataset
from operator import itemgetter,attrgetter
from functools import partial,wraps
from torch.optim import lr_scheduler
from torch import optim
from torchvision.io import read_image,ImageReadMode

from rapidai.datasets import *
from rapidai.conv import *
from rapidai.learner import *
from rapidai.activations import *
from rapidai.init import *
from rapidai.sgd import *
from rapidai.resnet import *
from rapidai.augment import *
from rapidai.accel import *
from rapidai.training import *




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_datasets.ipynb.

# %% ../nbs/02_datasets.ipynb 1
from __future__ import annotations
import math,numpy as np,matplotlib.pyplot as plt
from operator import itemgetter
from itertools import zip_longest
import fastcore.all as fc

from torch.utils.data import default_collate

from .training import *

# %% auto 0
__all__ = ['inplace', 'collate_dict', 'show_image', 'subplots', 'get_grid', 'show_images', 'DataLoaders']

# %% ../nbs/02_datasets.ipynb 23
def inplace(f):
    def _f(b):
        f(b)
        return b
    return _f

# %% ../nbs/02_datasets.ipynb 33
def collate_dict(ds):
    get = itemgetter(*ds.features)
    def _f(b): return get(default_collate(b))
    return _f

# %% ../nbs/02_datasets.ipynb 37
@fc.delegates(plt.Axes.imshow)
def show_image(im, ax=None, figsize=None, title=None, noframe=True, **kwargs):
    "Show a PIL or PyTorch image on `ax`."
    if fc.hasattrs(im, ('cpu','permute','detach')):
        im = im.detach().cpu()
        if len(im.shape)==3 and im.shape[0]<5: im=im.permute(1,2,0)
    elif not isinstance(im,np.ndarray): im=np.array(im)
    if im.shape[-1]==1: im=im[...,0]
    if ax is None: _,ax = plt.subplots(figsize=figsize)
    ax.imshow(im, **kwargs)
    if title is not None: ax.set_title(title)
    ax.set_xticks([]) 
    ax.set_yticks([]) 
    if noframe: ax.axis('off')
    return ax

# %% ../nbs/02_datasets.ipynb 41
@fc.delegates(plt.subplots, keep=True)
def subplots(
    nrows:int=1, # Number of rows in returned axes grid
    ncols:int=1, # Number of columns in returned axes grid
    figsize:tuple=None, # Width, height in inches of the returned figure
    imsize:int=3, # Size (in inches) of images that will be displayed in the returned figure
    suptitle:str=None, # Title to be set to returned figure
    **kwargs
): # fig and axs
    "A figure and set of subplots to display images of `imsize` inches"
    if figsize is None: figsize=(ncols*imsize, nrows*imsize)
    fig,ax = plt.subplots(nrows, ncols, figsize=figsize, **kwargs)
    if suptitle is not None: fig.suptitle(suptitle)
    if nrows*ncols==1: ax = np.array([ax])
    return fig,ax

# %% ../nbs/02_datasets.ipynb 45
@fc.delegates(subplots)
def get_grid(
    n:int, # Number of axes
    nrows:int=None, # Number of rows, defaulting to `int(math.sqrt(n))`
    ncols:int=None, # Number of columns, defaulting to `ceil(n/rows)`
    title:str=None, # If passed, title set to the figure
    weight:str='bold', # Title font weight
    size:int=14, # Title font size
    **kwargs,
): # fig and axs
    "Return a grid of `n` axes, `rows` by `cols`"
    if nrows: ncols = ncols or int(np.floor(n/nrows))
    elif ncols: nrows = nrows or int(np.ceil(n/ncols))
    else:
        nrows = int(math.sqrt(n))
        ncols = int(np.floor(n/nrows))
    fig,axs = subplots(nrows, ncols, **kwargs)
    for i in range(n, nrows*ncols): axs.flat[i].set_axis_off()
    if title is not None: fig.suptitle(title, weight=weight, size=size)
    return fig,axs

# %% ../nbs/02_datasets.ipynb 47
@fc.delegates(subplots)
def show_images(ims:list, # Images to show
                nrows:int|None=None, # Number of rows in grid
                ncols:int|None=None, # Number of columns in grid (auto-calculated if None)
                titles:list|None=None, # Optional list of titles for each image
                **kwargs):
    "Show all images `ims` as subplots with `rows` using `titles`"
    axs = get_grid(len(ims), nrows, ncols, **kwargs)[1].flat
    for im,t,ax in zip_longest(ims, titles or [], axs): show_image(im, ax=ax, title=t)

# %% ../nbs/02_datasets.ipynb 51
class DataLoaders:
    def __init__(self, *dls): self.train,self.valid = dls[:2]

    @classmethod
    def from_dd(cls, dd, batch_size, as_tuple=True, **kwargs):
        f = collate_dict(dd['train'])
        return cls(*get_dls(*dd.values(), bs=batch_size, collate_fn=f, **kwargs))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/12_diffusion-attn-cond.ipynb.

# %% auto 0
__all__ = ['abar', 'inv_abar', 'noisify', 'collate_ddpm', 'dl_ddpm', 'timestep_embedding', 'pre_conv', 'upsample', 'lin',
           'SelfAttention', 'SelfAttention2D', 'EmbResBlock', 'saved', 'DownBlock', 'UpBlock', 'EmbUNetModel',
           'ddim_step', 'sample', 'cond_sample']

# %% ../nbs/12_diffusion-attn-cond.ipynb 3
from .imports import *

from einops import rearrange
from fastprogress import progress_bar

# %% ../nbs/12_diffusion-attn-cond.ipynb 6
def abar(t): return (t*math.pi/2).cos()**2
def inv_abar(x): return x.sqrt().acos()*2/math.pi

def noisify(x0):
    device = x0.device
    n = len(x0)
    t = torch.rand(n,).to(x0).clamp(0,0.999)
    ε = torch.randn(x0.shape, device=device)
    abar_t = abar(t).reshape(-1, 1, 1, 1).to(device)
    xt = abar_t.sqrt()*x0 + (1-abar_t).sqrt()*ε
    return (xt, t.to(device)), ε

def collate_ddpm(b): return noisify(default_collate(b)[xl])
def dl_ddpm(ds): return DataLoader(ds, batch_size=bs, collate_fn=collate_ddpm, num_workers=4)

# %% ../nbs/12_diffusion-attn-cond.ipynb 10
def timestep_embedding(tsteps, emb_dim, max_period= 10000):
    exponent = -math.log(max_period) * torch.linspace(0, 1, emb_dim//2, device=tsteps.device)
    emb = tsteps[:,None].float() * exponent.exp()[None,:]
    emb = torch.cat([emb.sin(), emb.cos()], dim=-1)
    return F.pad(emb, (0,1,0,0)) if emb_dim%2==1 else emb

# %% ../nbs/12_diffusion-attn-cond.ipynb 11
def pre_conv(ni, nf, ks=3, stride=1, act=nn.SiLU, norm=None, bias=True):
    layers = nn.Sequential()
    if norm: layers.append(norm(ni))
    if act : layers.append(act())
    layers.append(nn.Conv2d(ni, nf, stride=stride, kernel_size=ks, padding=ks//2, bias=bias))
    return layers

# %% ../nbs/12_diffusion-attn-cond.ipynb 12
def upsample(nf): return nn.Sequential(nn.Upsample(scale_factor=2.), nn.Conv2d(nf, nf, 3, padding=1))

# %% ../nbs/12_diffusion-attn-cond.ipynb 13
def lin(ni, nf, act=nn.SiLU, norm=None, bias=True):
    layers = nn.Sequential()
    if norm: layers.append(norm(ni))
    if act : layers.append(act())
    layers.append(nn.Linear(ni, nf, bias=bias))
    return layers

# %% ../nbs/12_diffusion-attn-cond.ipynb 15
class SelfAttention(nn.Module):
    def __init__(self, ni, attn_chans, transpose=True):
        super().__init__()
        self.nheads = ni//attn_chans
        self.scale = math.sqrt(ni/self.nheads)
        self.norm = nn.LayerNorm(ni)
        self.qkv = nn.Linear(ni, ni*3)
        self.proj = nn.Linear(ni, ni)
        self.t = transpose
    
    def forward(self, x):
        n,c,s = x.shape
        if self.t: x = x.transpose(1, 2)
        x = self.norm(x)
        x = self.qkv(x)
        x = rearrange(x, 'n s (h d) -> (n h) s d', h=self.nheads)
        q,k,v = torch.chunk(x, 3, dim=-1)
        s = (q@k.transpose(1,2))/self.scale
        x = s.softmax(dim=-1)@v
        x = rearrange(x, '(n h) s d -> n s (h d)', h=self.nheads)
        x = self.proj(x)
        if self.t: x = x.transpose(1, 2)
        return x

# %% ../nbs/12_diffusion-attn-cond.ipynb 16
class SelfAttention2D(SelfAttention):
    def forward(self, x):
        n,c,h,w = x.shape
        return super().forward(x.view(n, c, -1)).reshape(n,c,h,w)

# %% ../nbs/12_diffusion-attn-cond.ipynb 17
class EmbResBlock(nn.Module):
    def __init__(self, n_emb, ni, nf=None, ks=3, act=nn.SiLU, norm=nn.BatchNorm2d, attn_chans=0):
        super().__init__()
        if nf is None: nf = ni
        self.emb_proj = nn.Linear(n_emb, nf*2)
        self.conv1 = pre_conv(ni, nf, ks, act=act, norm=norm)
        self.conv2 = pre_conv(nf, nf, ks, act=act, norm=norm)
        self.idconv = fc.noop if ni==nf else nn.Conv2d(ni, nf, 1)
        self.attn = False
        if attn_chans: self.attn = SelfAttention2D(nf, attn_chans)

    def forward(self, x, t):
        inp = x
        x = self.conv1(x)
        emb = self.emb_proj(F.silu(t))[:, :, None, None]
        scale,shift = torch.chunk(emb, 2, dim=1)
        x = x*(1+scale) + shift
        x = self.conv2(x)
        x = x + self.idconv(inp)
        if self.attn: x = x + self.attn(x)
        return x

# %% ../nbs/12_diffusion-attn-cond.ipynb 18
def saved(m, blk):
    m_ = m.forward

    @wraps(m.forward)
    def _f(*args, **kwargs):
        res = m_(*args, **kwargs)
        blk.saved.append(res)
        return res

    m.forward = _f
    return m

# %% ../nbs/12_diffusion-attn-cond.ipynb 19
class DownBlock(nn.Module):
    def __init__(self, n_emb, ni, nf, add_down=True, num_layers=1, attn_chans=0):
        super().__init__()
        self.resnets = nn.ModuleList([saved(EmbResBlock(n_emb, ni if i==0 else nf, nf, attn_chans=attn_chans), self)
                                      for i in range(num_layers)])
        self.down = saved(nn.Conv2d(nf, nf, 3, stride=2, padding=1), self) if add_down else nn.Identity()

    def forward(self, x, t):
        self.saved = []
        for resnet in self.resnets: x = resnet(x, t)
        x = self.down(x)
        return x

# %% ../nbs/12_diffusion-attn-cond.ipynb 20
class UpBlock(nn.Module):
    def __init__(self, n_emb, ni, prev_nf, nf, add_up=True, num_layers=2, attn_chans=0):
        super().__init__()
        self.resnets = nn.ModuleList(
            [EmbResBlock(n_emb, (prev_nf if i==0 else nf)+(ni if (i==num_layers-1) else nf), nf, attn_chans=attn_chans)
            for i in range(num_layers)])
        self.up = upsample(nf) if add_up else nn.Identity()

    def forward(self, x, t, ups):
        for resnet in self.resnets: x = resnet(torch.cat([x, ups.pop()], dim=1), t)
        return self.up(x)

# %% ../nbs/12_diffusion-attn-cond.ipynb 21
class EmbUNetModel(nn.Module):
    def __init__( self, in_channels=3, out_channels=3, nfs=(224,448,672,896), num_layers=1, attn_chans=8, attn_start=1):
        super().__init__()
        self.conv_in = nn.Conv2d(in_channels, nfs[0], kernel_size=3, padding=1)
        self.n_temb = nf = nfs[0]
        n_emb = nf*4
        self.emb_mlp = nn.Sequential(lin(self.n_temb, n_emb, norm=nn.BatchNorm1d),
                                     lin(n_emb, n_emb))
        self.downs = nn.ModuleList()
        n = len(nfs)
        for i in range(n):
            ni = nf
            nf = nfs[i]
            self.downs.append(DownBlock(n_emb, ni, nf, add_down=i!=n-1, num_layers=num_layers,
                                        attn_chans=0 if i<attn_start else attn_chans))
        self.mid_block = EmbResBlock(n_emb, nfs[-1])

        rev_nfs = list(reversed(nfs))
        nf = rev_nfs[0]
        self.ups = nn.ModuleList()
        for i in range(n):
            prev_nf = nf
            nf = rev_nfs[i]
            ni = rev_nfs[min(i+1, len(nfs)-1)]
            self.ups.append(UpBlock(n_emb, ni, prev_nf, nf, add_up=i!=n-1, num_layers=num_layers+1,
                                    attn_chans=0 if i>=n-attn_start else attn_chans))
        self.conv_out = pre_conv(nfs[0], out_channels, act=nn.SiLU, norm=nn.BatchNorm2d, bias=False)

    def forward(self, inp):
        x,t = inp
        temb = timestep_embedding(t, self.n_temb)
        emb = self.emb_mlp(temb)
        x = self.conv_in(x)
        saved = [x]
        for block in self.downs: x = block(x, emb)
        saved += [p for o in self.downs for p in o.saved]
        x = self.mid_block(x, emb)
        for block in self.ups: x = block(x, emb, saved)
        return self.conv_out(x)

# %% ../nbs/12_diffusion-attn-cond.ipynb 28
def ddim_step(x_t, noise, abar_t, abar_t1, bbar_t, bbar_t1, eta, sig, clamp=True):
    sig = ((bbar_t1/bbar_t).sqrt() * (1-abar_t/abar_t1).sqrt()) * eta
    x_0_hat = ((x_t-(1-abar_t).sqrt()*noise) / abar_t.sqrt())
    if clamp: x_0_hat = x_0_hat.clamp(-1,1)
    if bbar_t1<=sig**2+0.01: sig=0.  # set to zero if very small or NaN
    x_t = abar_t1.sqrt()*x_0_hat + (bbar_t1-sig**2).sqrt()*noise
    x_t += sig * torch.randn(x_t.shape).to(x_t)
    return x_0_hat,x_t

# %% ../nbs/12_diffusion-attn-cond.ipynb 29
@torch.no_grad()
def sample(f, model, sz, steps, eta=1., clamp=True):
    model.eval()
    ts = torch.linspace(1-1/steps,0,steps)
    x_t = torch.randn(sz).cuda()
    preds = []
    for i,t in enumerate(progress_bar(ts)):
        t = t[None].cuda()
        abar_t = abar(t)
        noise = model((x_t, t))
        abar_t1 = abar(t-1/steps) if t>=1/steps else torch.tensor(1)
        x_0_hat,x_t = f(x_t, noise, abar_t, abar_t1, 1-abar_t, 1-abar_t1, eta, 1-((i+1)/100), clamp=clamp)
        preds.append(x_0_hat.float().cpu())
    return preds

# %% ../nbs/12_diffusion-attn-cond.ipynb 43
@torch.no_grad()
def cond_sample(c, f, model, sz, steps, eta=1.):
    ts = torch.linspace(1-1/steps,0,steps)
    x_t = torch.randn(sz).cuda()
    c = x_t.new_full((sz[0],), c, dtype=torch.int32)
    preds = []
    for i,t in enumerate(progress_bar(ts)):
        t = t[None].cuda()
        abar_t = abar(t)
        noise = model((x_t, t, c))
        abar_t1 = abar(t-1/steps) if t>=1/steps else torch.tensor(1)
        x_0_hat,x_t = f(x_t, noise, abar_t, abar_t1, 1-abar_t, 1-abar_t1, eta, 1-((i+1)/100))
        preds.append(x_0_hat.float().cpu())
    return preds



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_convolutions.ipynb.

# %% auto 0
__all__ = ['def_device', 'conv', 'device', 'to_device', 'collate_device']

# %% ../nbs/03_convolutions.ipynb 2
import torch
from torch import nn

from torch.utils.data import default_collate
from typing import Mapping

from .training import *
from .datasets import *

# %% ../nbs/03_convolutions.ipynb 69
def conv(ni, nf, ks=3, stride=2, act=True):
    res = nn.Conv2d(ni, nf, stride=stride, kernel_size=ks, padding=ks//2)
    if act: res = nn.Sequential(res, nn.ReLU())
    return res

# %% ../nbs/03_convolutions.ipynb 74
def_device = 'mps' if torch.backends.mps.is_available() else 'cuda' if torch.cuda.is_available() else 'cpu'

def device():
    return 'Accelerator: Apple-GPU' if torch.backends.mps.is_available() else 'Accelerator: GPU' if torch.cuda.is_available() else 'No accelerator found using cpu'
 
 
def to_device(x, device=def_device):
    if isinstance(x, torch.Tensor): return x.to(device)
    if isinstance(x, Mapping): return {k:v.to(device) for k,v in x.items()}
    return type(x)(to_device(o, device) for o in x)

def collate_device(b): return to_device(default_collate(b))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_activations.ipynb.

# %% ../nbs/05_activations.ipynb 2
from __future__ import annotations
import random,math,torch,numpy as np,matplotlib.pyplot as plt
import fastcore.all as fc
from functools import partial

from .datasets import *
from .learner import *

# %% auto 0
__all__ = ['set_seed', 'Hook', 'Hooks', 'HooksCallback', 'append_stats', 'get_hist', 'get_min', 'ActivationStats']

# %% ../nbs/05_activations.ipynb 4
def set_seed(seed, deterministic=False):
    torch.use_deterministic_algorithms(deterministic)
    torch.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)

# %% ../nbs/05_activations.ipynb 30
class Hook():
    def __init__(self, m, f): self.hook = m.register_forward_hook(partial(f, self))
    def remove(self): self.hook.remove()
    def __del__(self): self.remove()

# %% ../nbs/05_activations.ipynb 42
class Hooks(list):
    def __init__(self, ms, f): super().__init__([Hook(m, f) for m in ms])
    def __enter__(self, *args): return self
    def __exit__ (self, *args): self.remove()
    def __del__(self): self.remove()
    def __delitem__(self, i):
        self[i].remove()
        super().__delitem__(i)
    def remove(self):
        for h in self: h.remove()

# %% ../nbs/05_activations.ipynb 46
class HooksCallback(Callback):
    def __init__(self, hookfunc, mod_filter=fc.noop, on_train=True, on_valid=False, mods=None):
        fc.store_attr()
        super().__init__()
    
    def before_fit(self, learn):
        if self.mods: mods=self.mods
        else: mods = fc.filter_ex(learn.model.modules(), self.mod_filter)
        self.hooks = Hooks(mods, partial(self._hookfunc, learn))

    def _hookfunc(self, learn, *args, **kwargs):
        if (self.on_train and learn.training) or (self.on_valid and not learn.training): self.hookfunc(*args, **kwargs)

    def after_fit(self, learn): self.hooks.remove()
    def __iter__(self): return iter(self.hooks)
    def __len__(self): return len(self.hooks)

# %% ../nbs/05_activations.ipynb 51
def append_stats(hook, mod, inp, outp):
    if not hasattr(hook,'stats'): hook.stats = ([],[],[])
    acts = to_cpu(outp)
    hook.stats[0].append(acts.mean())
    hook.stats[1].append(acts.std())
    hook.stats[2].append(acts.abs().histc(40,0,10))

# %% ../nbs/05_activations.ipynb 53
# Thanks to @ste for initial version of histgram plotting code
def get_hist(h): return torch.stack(h.stats[2]).t().float().log1p()

# %% ../nbs/05_activations.ipynb 55
def get_min(h):
    h1 = torch.stack(h.stats[2]).t().float()
    return h1[0]/h1.sum(0)

# %% ../nbs/05_activations.ipynb 58
class ActivationStats(HooksCallback):
    def __init__(self, mod_filter=fc.noop): super().__init__(append_stats, mod_filter)

    def color_dim(self, figsize=(11,5)):
        fig,axes = get_grid(len(self), figsize=figsize)
        for ax,h in zip(axes.flat, self):
            show_image(get_hist(h), ax, origin='lower')

    def dead_chart(self, figsize=(11,5)):
        fig,axes = get_grid(len(self), figsize=figsize)
        for ax,h in zip(axes.flatten(), self):
            ax.plot(get_min(h))
            ax.set_ylim(0,1)

    def plot_stats(self, figsize=(10,4)):
        fig,axs = plt.subplots(1,2, figsize=figsize)
        for h in self:
            for i in 0,1: axs[i].plot(h.stats[i])
        axs[0].set_title('Means')
        axs[1].set_title('Stdevs')
        plt.legend(fc.L.range(self))
